import re
from dataclasses import dataclass, field

import click

from aox.styling.shortcuts import e_error
from aox.web.aoc import WebAoc


__all__ = ['AccountScraper']


@dataclass
class AccountScraper:
    web_aoc: WebAoc = field(default_factory=WebAoc)

    def collect_data(self):
        events_page = self.web_aoc.get_events_page()
        if not events_page:
            return None

        username = self.get_username(events_page)
        if not username:
            return None

        total_stars = self.get_total_stars(events_page)

        collected_data = {
            "username": username,
            "total_stars": total_stars,
            "years": {},
        }

        for year, stars in self.get_years_stars(events_page).items():
            year_data = self.collect_year(year, stars)
            if year_data is None:
                continue
            collected_data["years"][year] = year_data
        return collected_data

    def get_username(self, events_page):
        """
        Extract the username. If there isn't a username, there is no information
        to gather.

        >>> import bs4
        >>> AccountScraper(WebAoc('test-session')).get_username(
        ...     bs4.BeautifulSoup(
        ...         "<html><body>"
        ...         "<div class='user'>Test User <span class='star-count'>50*",
        ...         "html.parser"))
        'Test User'
        >>> AccountScraper(WebAoc('test-session')).get_username(
        ...     bs4.BeautifulSoup(
        ...         "<html><body>"
        ...         "<a href='/2020/auth/login'>[Log In]</a>",
        ...         "html.parser"))
        Either ... user name
        """
        user_nodes = events_page.select(".user")
        if not user_nodes:
            click.echo(
                f"Either the session ID in {e_error('AOC_SESSION_ID')} is "
                f"wrong, or it has expired: could not find the user name")
            return None

        user_node = user_nodes[0]
        text_children = [
            child
            for child in user_node.children
            if isinstance(child, str)
        ]
        if not text_children:
            click.echo(
                f"Either the user name is blank or the format has changed")
            return None
        return text_children[0].strip()

    def get_total_stars(self, events_page):
        """
        Grab the total stars

        >>> import bs4
        >>> AccountScraper(WebAoc('test-session')).get_total_stars(
        ...     bs4.BeautifulSoup(
        ...         "<html><body>"
        ...         "<div class='eventlist-event'><a href='/'>[2020]</a>"
        ...         "<span class='star-count'>39*</span></div>"
        ...         "<p>Total stars: <span class='star-count'>64*</span>",
        ...         "html.parser"))
        64
        >>> AccountScraper(WebAoc('test-session')).get_total_stars(
        ...     bs4.BeautifulSoup(
        ...         "<html><body>"
        ...         "<div class='eventlist-event'><a href='/'>[2020]</a>"
        ...         "<span class='star-count'>39*</span></div>",
        ...         "html.parser"))
        """
        total_stars_nodes = events_page.select("p > .star-count")
        return self.parse_star_count(total_stars_nodes)

    def parse_star_count(self, stars_nodes, default=None):
        """
        Star counts are represented in the same way everywhere

        >>> import bs4
        >>> AccountScraper(WebAoc('test-session')).parse_star_count(
        ...     [bs4.BeautifulSoup(
        ...         "<span class='star-count'>39*</span></div>",
        ...         "html.parser")])
        39
        >>> AccountScraper(WebAoc('test-session')).parse_star_count(
        ...     [bs4.BeautifulSoup(
        ...         "<span class='star-count'>39*</span></div>",
        ...         "html.parser")], default=0)
        39
        >>> AccountScraper(WebAoc('test-session')).parse_star_count([])
        >>> AccountScraper(WebAoc('test-session')).parse_star_count([], 0)
        0
        """
        if not stars_nodes:
            return default
        stars_node = stars_nodes[0]

        return self.parse_star_count_text(stars_node.text, default=default)

    re_stars = re.compile(r'^(\d+)\*$')

    def parse_star_count_text(self, text, default=None):
        """
        >>> AccountScraper(WebAoc('test-session')).parse_star_count_text('0*')
        0
        >>> AccountScraper(WebAoc('test-session')).parse_star_count_text('120*')
        120
        >>> AccountScraper(WebAoc('test-session'))\\
        ...     .parse_star_count_text('120*', default=0)
        120
        >>> AccountScraper(WebAoc('test-session'))\\
        ...     .parse_star_count_text('star?')
        >>> AccountScraper(WebAoc('test-session'))\\
        ...     .parse_star_count_text('star?', default=0)
        0
        """
        stars_match = self.re_stars.match(text.strip())
        if not stars_match:
            return default

        stars_text, = stars_match.groups()
        return int(stars_text)

    def get_years_stars(self, events_page):
        """
        >>> import bs4
        >>> AccountScraper(WebAoc('test-session')).get_years_stars(
        ...     bs4.BeautifulSoup(
        ...         "<html><body>"
        ...         "<div class='eventlist-event'><a href='/'>[2020]</a>"
        ...         "<span class='star-count'>39*</span></div>"
        ...         "<div class='eventlist-event'><a href='/'>[2019]</a>"
        ...         "<span class='star-count'>20*</span></div>"
        ...         "<div class='eventlist-event'><a href='/'>[2018]</a>"
        ...         " </div>"
        ...         "<p>Total stars: <span class='star-count'>64*</span>",
        ...         "html.parser"))
        {2020: 39, 2019: 20}
        >>> AccountScraper(WebAoc('test-session')).get_years_stars(
        ...     bs4.BeautifulSoup(
        ...         "<html><body>"
        ...         "<p>Total stars: <span class='star-count'>64*</span>",
        ...         "html.parser"))
        {}
        """
        stars_nodes = events_page.select(".eventlist-event .star-count")
        years_nodes = [node.parent for node in stars_nodes]
        return {
            year: stars
            for year, stars
            in filter(None, map(self.get_year_stars, years_nodes))
            if stars
        }

    re_year = re.compile(r'^\[(\d+)]$')

    def get_year_stars(self, year_node):
        """
        Get the stars for a year from a year node in the events page

        >>> import bs4
        >>> AccountScraper(WebAoc('test-session')).get_year_stars(
        ...     bs4.BeautifulSoup(
        ...         "<div class='eventlist-event'><a href='/'>[2020]</a>"
        ...         "<span class='star-count'>39*</span></div>",
        ...         "html.parser"))
        (2020, 39)
        >>> AccountScraper(WebAoc('test-session')).get_year_stars(
        ...     bs4.BeautifulSoup(
        ...         "<div class='eventlist-event'><a href='/'>[2020]</a>"
        ...         " </div>",
        ...         "html.parser"))
        (2020, 0)
        """
        year_name_node = year_node.findChild('a')
        if not year_name_node:
            return None
        year_name_match = self.re_year.match(year_name_node.text)
        if not year_name_match:
            return None
        year_text, = year_name_match.groups()
        year = int(year_text)

        stars_nodes = year_node.select('.star-count')
        stars = self.parse_star_count(stars_nodes, default=0)

        return year, stars

    def collect_year(self, year, stars):
        year_page = self.web_aoc.get_year_page(year)
        if year_page is None:
            return None

        return self.collect_year_from_page(year_page, year, stars)

    def collect_year_from_page(self, year_page, year, stars):
        """
        >>> import bs4
        >>> AccountScraper(WebAoc('test-session')).collect_year_from_page(
        ...     bs4.BeautifulSoup(
        ...         "<html><body><article>"
        ...         "<pre class='calendar calendar-perfect'>"
        ...         "<a aria-label='Day 15, two stars' href='/2020/day/15' "
        ...         "class='calendar-day15 calendar-verycomplete'>"
        ...         "Very fancy day text"
        ...         "<span class='calendar-day'>15</span> <span "
        ...         "class='calendar-mark-complete'>*</span>"
        ...         "<span class='calendar-mark-verycomplete'>*</span></a>",
        ...         "html.parser"), 2020, 38)
        {'year': 2020, 'stars': 38, 'days': {15: 2}}
        """
        year_data = {
            "year": year,
            "stars": stars,
            "days": {},
        }

        days_nodes = year_page.select(
            '.calendar > a[class^="calendar-day"]')
        for day_node in days_nodes:
            day, stars = self.collect_day(day_node)
            if (day, stars) == (None, None):
                continue
            year_data["days"][day] = stars

        return year_data

    def collect_day(self, day_node):
        """
        >>> import bs4
        >>> AccountScraper(WebAoc('test-session')).collect_day(
        ...     bs4.BeautifulSoup(
        ...         "<a aria-label='Day 15, two stars' href='/2020/day/15' "
        ...         "class='calendar-day15 calendar-verycomplete'>"
        ...         "Very fancy day text"
        ...         "<span class='calendar-day'>15</span> <span "
        ...         "class='calendar-mark-complete'>*</span>"
        ...         "<span class='calendar-mark-verycomplete'>*</span></a>",
        ...         "html.parser").a)
        (15, 2)
        >>> AccountScraper(WebAoc('test-session')).collect_day(
        ...     bs4.BeautifulSoup(
        ...         "<a aria-label='Day 15, two stars' href='/2020/day/15' "
        ...         "class='calendar-day15 calendar-complete'>"
        ...         "Very fancy day text"
        ...         "<span class='calendar-day'>15</span> <span "
        ...         "class='calendar-mark-complete'>*</span>"
        ...         "<span class='calendar-mark-verycomplete'>*</span></a>",
        ...         "html.parser").a)
        (15, 1)
        >>> AccountScraper(WebAoc('test-session')).collect_day(
        ...     bs4.BeautifulSoup(
        ...         "<a aria-label='Day 15, two stars' href='/2020/day/15' "
        ...         "class='calendar-day15 '>"
        ...         "Very fancy day text"
        ...         "<span class='calendar-day'>15</span> <span "
        ...         "class='calendar-mark-complete'>*</span>"
        ...         "<span class='calendar-mark-verycomplete'>*</span></a>",
        ...         "html.parser").a)
        (15, 0)
        >>> AccountScraper(WebAoc('test-session')).collect_day(
        ...     bs4.BeautifulSoup(
        ...         "<a aria-label='Day 15, two stars' href='/2020/day/15' "
        ...         "class='calendar-day15 '>"
        ...         "Very fancy day text"
        ...         "<span class='what-calendar-day'>15</span> <span "
        ...         "class='calendar-mark-complete'>*</span>"
        ...         "<span class='calendar-mark-verycomplete'>*</span></a>",
        ...         "html.parser").a)
        (None, None)
        """
        day_name_nodes = day_node.select(".calendar-day")
        if not day_name_nodes:
            return None, None
        day_name_node = day_name_nodes[0]
        day_text = day_name_node.text
        try:
            day = int(day_text)
        except ValueError:
            return None, None

        class_names = day_node.attrs['class']
        if 'calendar-verycomplete' in class_names:
            stars = 2
        elif 'calendar-complete' in class_names:
            stars = 1
        else:
            stars = 0

        return day, stars



import inspect
import os
from pathlib import Path


__all__ = [
    'get_caller_file_name',
    'get_root_directory',
    'get_current_directory',
]


def get_caller_file_name(skip_frames=1):
    """
    Get the file name of the original caller of the calling function.

    :param skip_frames: The number of frames to skip to get the caller. If you
    want your filename, set to 0, if you want your caller's leave to default or
    set to 1, and any number above that to traverse your callers.

    >>> get_caller_file_name(0)
    '.../utils/paths.py'
    >>> def example_caller():
    ...     return get_caller_file_name()
    >>> example_caller()
    '.../utils/paths.py'
    """
    caller = inspect.stack()[1 + skip_frames]
    caller_frame, *_ = caller
    caller_globals = caller_frame.f_globals
    module_file_name = caller_globals.get('__file__', None)

    return module_file_name


def get_root_directory():
    """
    Get the root directory of this repo

    >>> import sys
    >>> if sys.version_info >= (3, 9):
    ...     get_current_directory().is_relative_to(get_root_directory())
    ... else:
    ...     True
    True
    >>> if sys.version_info >= (3, 9):
    ...     str(get_current_directory().relative_to(get_root_directory()))
    ... else:
    ...     'aox/utils'
    'aox/utils'
    """
    return Path('.').absolute()


def get_current_directory(module_file_name=None):
    """
    Get the directory of the module where this is called from. Usually called as
    `get_current_directory()`, or `get_current_directory(__file__)`.

    >>> str(get_current_directory())
    '.../utils'
    """
    if module_file_name is None:
        module_file_name = get_caller_file_name()
        if not module_file_name:
            raise Exception(
                f"Could not get the file name of the calling module "
                f"automatically - call with `__file__` as the only parameter, "
                f"or specify the file name explicitly.")
    return Path(os.path.dirname(os.path.realpath(module_file_name)))



from enum import Enum


__all__ = ['StringEnum']


class StringEnum(Enum):
    """
    An enum class that auto-assigns values to be the same as names

    >>> from enum import auto
    >>> class TestEnum(StringEnum):
    ...     Foo = auto()
    ...     ClassName = auto()
    ...     ClassNameWithAWSUsingSDK = auto()
    >>> TestEnum.Foo.value
    'foo'
    >>> TestEnum.ClassName.value
    'class-name'
    >>> TestEnum.ClassNameWithAWSUsingSDK.value
    'class-name-with-aws-using-sdk'
    """
    # noinspection PyMethodParameters
    def _generate_next_value_(name, start, count, last_values):
        from aox.utils import tokenize_camel_case
        return tokenize_camel_case(name)



import importlib.util
import os
from pathlib import Path
from typing import Union, Optional

__all__ = [
    'load_module_from_path',
    'try_import_module',
    'is_import_error_for_own_module',
]

from aox.utils.typing_utils import Module


def load_module_from_path(path: Union[Path, str], module_name: str = None):
    """
    Load a module only by it's given path. This is useful when you want to load
    a single Python module from a path that either isn't in the `PYTHONPATH`, or
    isn't a valid Python path.

    Currently used for loading sensitive user settings in
    `.aox/sensitive_user_settings.py`.

    :param path: The full path to the module
    :param module_name: The name to give to the loaded module. By default it
    uses the filename
    :return:
    """
    if isinstance(path, str):
        path = Path(path)
    if module_name is None:
        module_name, _ = os.path.splitext(path.name)
    spec = importlib.util.spec_from_file_location(module_name, path)
    module = importlib.util.module_from_spec(spec)
    # noinspection PyUnresolvedReferences
    spec.loader.exec_module(module)

    return module


def try_import_module(module_name: str) -> Optional[Module]:
    """
    Try to import a module, and return nothing if it doesn't exit

    >>> try_import_module('aox.utils.non_existing_module')
    >>> try_import_module('aox.utils.module_utils')
    <module ...>
    """
    try:
        return importlib.import_module(module_name)
    except ImportError as e:
        if not is_import_error_for_own_module(module_name, e):
            raise

        return None


def is_import_error_for_own_module(
        module_name: str, import_error: ImportError) -> bool:
    """
    Find if an `ImportError` is due to the file not existing, or if it is due to
    some other issue, eg cyclical import.
    """
    parts = module_name.split('.')
    partial_module_names = [
        '.'.join(parts[:last])
        for last in range(1, len(parts) + 1)
    ]
    acceptable_messages = {
        f"No module named '{partial_module_name}'"
        for partial_module_name in partial_module_names
    }
    error_message = getattr(import_error, 'msg')
    return error_message in acceptable_messages



from typing import Optional

from aox.settings.settings_class import Settings


__all__ = ['UninitialisedSettingsError', 'SettingsProxy', 'settings_proxy']


class UninitialisedSettingsError(Exception):
    """
    Error signifying that the settings were trying to be accessed before they
    were initialised.
    """

    DEFAULT_MESSAGE = "The settings have not been initialised yet"

    def __init__(self, message=DEFAULT_MESSAGE, *args):
        super().__init__(message, *args)


class SettingsProxy:
    """A proxy to safely access the current active settings instance"""
    settings: Optional['Settings'] = None
    """The actual settings instance"""

    def __call__(self, raise_if_missing: bool = True) -> Optional['Settings']:
        """Shortcut for `SettingsProxy.get`"""
        return self.get(raise_if_missing)

    def get(self, raise_if_missing: bool = True) -> Optional['Settings']:
        """Get the current (initialised or not) settings"""
        if raise_if_missing and self.settings is None:
            raise UninitialisedSettingsError()
        return self.settings

    def ensure_default(self) -> Optional['Settings']:
        """Set the default settings, if not initialised yet"""
        if self.has():
            return
        return self.set_default()

    def set_default(self) -> Optional['Settings']:
        """Update the settings instance to the one gotten by default"""
        return self.set(Settings.from_default())

    def set(self, new_settings: Optional['Settings']) -> Optional['Settings']:
        """Update the settings instance to custom one"""
        self.settings = new_settings

        return self.settings

    def has(self) -> bool:
        """Check whether settings have been initialised"""
        return self.settings is not None


settings_proxy = SettingsProxy()



from typing import TypeVar, Iterable, Tuple

__all__ = ['in_groups']

T = TypeVar('T')


def in_groups(iterable: Iterable[T], size: int) -> Iterable[Tuple[T, ...]]:
    """
    >>> list(in_groups([], 3))
    []
    >>> list(in_groups(range(5), 3))
    [(0, 1, 2), (3, 4)]
    >>> # noinspection PyUnresolvedReferences
    >>> list(in_groups((x for x in range(5)), 3))
    [(0, 1, 2), (3, 4)]
    >>> # noinspection PyUnresolvedReferences
    >>> list(in_groups((x for x in range(6)), 3))
    [(0, 1, 2), (3, 4, 5)]
    >>> # noinspection PyUnresolvedReferences
    >>> list(in_groups((x for x in range(5)), 6))
    [(0, 1, 2, 3, 4)]
    >>> # noinspection PyUnresolvedReferences
    >>> list(in_groups((x for x in range(5)), 10))
    [(0, 1, 2, 3, 4)]
    """
    if size < 1:
        raise ValueError(f"Expected size larger than 0, got {size}")
    group = ()
    for item in iterable:
        group += (item,)
        if len(group) >= size:
            yield group
            group = ()
    if group:
        yield group



#!/usr/bin/env python
import os
from distutils.core import setup
from pathlib import Path

import setuptools

from aox.version import AOX_PACKAGE_VERSION_LABEL

current_directory = Path(os.path.dirname(os.path.realpath(__file__)))


setup(
    name='aox',
    version=AOX_PACKAGE_VERSION_LABEL,
    description='AdventOfCode Submissions Framework',
    author='Costas Basdekis',
    author_email='costas@basdekis.io',
    url='https://github.com/costas-basdekis/aox',
    long_description=current_directory.joinpath('README.md').read_text(),
    long_description_content_type='text/markdown',
    packages=setuptools.find_packages(include=("aox*",)),
    scripts=[
        'scripts/aox',
    ],
    license='MIT License',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Software Development',
        'Typing :: Typed',
    ],
)



"""
A class that provides an interface, and some boilerplate, for every challenge.
"""

import doctest
import importlib
import sys

from aox.settings import settings_proxy
from aox.utils import has_method_arguments


__all__ = ['PlayNotImplementedError', 'BaseChallenge']


class PlayNotImplementedError(NotImplementedError):
    pass


class BaseChallenge:
    """
    The base class for every challenge.

    The only method that really matters is `solve`, but it also allows for
    specifying a `play` method, that is normally used for interactive
    challenges, or debugging reasons.
    """
    part_a_for_testing = None
    optionflags = doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE

    def __init__(self):
        self.module = self.get_module()
        self.year, self.day, self.part = settings_proxy()\
            .challenges_boilerplate\
            .extract_from_filename(self.module.__file__)
        self.input = self.get_input()

    @classmethod
    def get_module(cls):
        return sys.modules[cls.__module__]

    @classmethod
    def is_main_module(cls):
        main_module = sys.modules.get('__main__')
        return cls.get_module() == main_module

    def get_input(self):
        """Get the input for the challenge"""
        return settings_proxy().challenges_boilerplate\
            .get_day_input_filename(self.year, self.day)\
            .read_text()

    @classmethod
    def main(cls, extra_args=None):
        """
        A way to run the functionality from the command line, but limited to
        only this specific challenge.
        """
        if not cls.is_main_module():
            return
        from aox.command_line.command import cli
        return cli(args=(cls.get_main_args(extra_args=extra_args)))

    @classmethod
    def get_main_args(cls, extra_args=None):
        """The CLI arguments to simulate an invocation of this challenge"""
        if extra_args is None:
            extra_args = sys.argv[1:]
        return [
            'challenge',
            '--path', cls.get_module().__file__,
            '0',
            '0',
            'a',
        ] + extra_args

    # noinspection PyUnusedLocal
    def default_solve(self, _input=None, debugger=None):
        """Convenient method to call `solve` with the input from the disk"""
        if _input is None:
            _input = self.input
        if debugger is None:
            from aox.challenge import Debugger
            debugger = Debugger(enabled=False)
        if has_method_arguments(self.solve, "debugger"):
            return self.solve(_input, debugger=debugger)
        else:
            # noinspection PyArgumentList
            return self.solve(_input, debug=debugger)

    def solve(self, _input, debugger):
        """
        The main method that needs to be overridden by every challenge.

        The solution must be return from this method, either as a number, or as
        a plain string.
        """
        raise NotImplementedError()

    def play(self):
        """
        Some challenges can benefit from providing an interactive or visual
        mode.
        """
        raise PlayNotImplementedError(f"Challenge has not implemented play")

    def get_test_modules(self):
        """
        Get the modules to test. Usually it's the current challenge, but
        sometimes part A also needs to be tested when testing part B, and also
        any other utility functionality not inside either part.
        """
        modules = [
            importlib.import_module(type(self).__module__),
        ]
        if self.part_a_for_testing:
            modules.append(self.part_a_for_testing)
        return modules



"""
Implements the filtering functionality that is enabled by the enhanced `testmod`
methods in :mod:`doctest_enhanced_testmod`.
"""

import copy
import doctest
from doctest import DocTest
from typing import List, Optional, Type, Union

from aox.testing.doctest_filter_parsing import IndexedDocTestFilterParser, \
    DocTestFilter, DocTestFilterParser


class FilteringDocTestFinder(doctest.DocTestFinder):
    def __init__(self, verbose=False, parser=doctest.DocTestParser(),
                 recurse=True, exclude_empty=True,
                 parser_class: Type[DocTestFilterParser] =
                 IndexedDocTestFilterParser,
                 filters_text=""):
        super().__init__(verbose=verbose, parser=parser, recurse=recurse,
                         exclude_empty=exclude_empty)
        self.filters: List[DocTestFilter] = \
            parser_class().parse_filters(filters_text)

    def find(self, obj, name=None, module=None, globs=None, extraglobs=None):
        tests = super().find(obj, name, module, globs, extraglobs)

        if self.filters:
            tests = self.filter_tests(tests, self.filters)

        return tests

    def filter_tests(self, tests: List[DocTest],
                     filters_or_text: Union[str, List[DocTestFilter]]) \
            -> List[DocTest]:
        """
        Filter collected tests according a user specification, by using
        `DocTestFilter`s.
        :param tests: The list of tests to filter
        :param filters_or_text: Either a user-specified string, or a parsed one
        :return: A (not necessarily strict) subset of the passed in tests
        """
        if isinstance(filters_or_text, str):
            filters_text = filters_or_text
            filters = DocTestFilterParser().parse_filters(filters_text)
        else:
            filters = filters_or_text

        return list(filter(None, (
            self.filter_test(filters, test)
            for test in tests
        )))

    def filter_test(self, filters, test: DocTest) -> Optional[DocTest]:
        """
        Filter a test according the user specifications. If a test shouldn't be
        run `None` is returned. If it should be, and all of its examples should
        as well, it's returned as is. If only a few of its examples should be,
        a new test with only the appropriate examples is returned.
        :param filters: The user-specified `DocTestFilter`s
        :param test: A test to either discard, accept it fully, or with a subset
        of its examples
        :return: Either `None` to discard the test, the test as is, or a copy
        with fewer examples
        """
        matching_filters = [
            _filter
            for _filter in filters
            if _filter.matches_test(test)
        ]
        if not matching_filters:
            return None
        examples = [
            example
            for index, example in enumerate(test.examples)
            if any(
                _filter.matches_example(test, example, index)
                for _filter in matching_filters
            )
        ]
        if not examples:
            return None

        if examples != test.examples:
            test = copy.copy(test)
            test.examples = examples

        return test



import re
from typing import Union


__all__ = ['tokenize_camel_case', 'add_thousands_separator']


RES_TOKENIZE_WORDS_BOUNDARY = [
    re.compile(r'([a-z])([A-Z])'),
    re.compile(r'([A-Z])([A-Z][a-z])'),
]


def tokenize_camel_case(name):
    """
    Covert a camel case name to a dash-separated token

    >>> tokenize_camel_case('abc')
    'abc'
    >>> tokenize_camel_case('functionName')
    'function-name'
    >>> tokenize_camel_case('ClassName')
    'class-name'
    >>> tokenize_camel_case('ClassNameWithAWSAbbreviationsSDK')
    'class-name-with-aws-abbreviations-sdk'
    """
    for re_tokenize_words_boundary in RES_TOKENIZE_WORDS_BOUNDARY:
        name = re_tokenize_words_boundary.sub(r'\1-\2', name)
    name = name.lower()

    return name


def add_thousands_separator(number: Union[int, float], separator: str = ",",
                            ) -> str:
    """
    >>> add_thousands_separator(0)
    '0'
    >>> add_thousands_separator(123)
    '123'
    >>> add_thousands_separator(1234)
    '1,234'
    >>> add_thousands_separator(12345)
    '12,345'
    >>> add_thousands_separator(123456)
    '123,456'
    >>> add_thousands_separator(12345678901234567890)
    '12,345,678,901,234,567,890'
    >>> add_thousands_separator(-12345.12345678)
    '-12,345.12345678'
    >>> add_thousands_separator(12345.12345678)
    '12,345.12345678'
    >>> add_thousands_separator(-12345)
    '-12,345'
    >>> add_thousands_separator(-12345.0)
    '-12,345'
    """
    reversed_digits = reversed(str(abs(int(number))))
    from aox.utils import in_groups
    reversed_groups = in_groups(reversed_digits, 3)
    groups = reversed(list(map(reversed, reversed_groups)))
    joined = separator.join(map("".join, groups))
    if number < 0:
        joined = f"-{joined}"
    if float(int(number)) != float(number):
        _, remainder = str(number).split(".")
        joined += f".{remainder}"

    return joined



"""
Provide shortcuts for styling parts of text when interacting with the console.
"""

import functools

import click


def make_colour(**colour_kwargs):
    """
    Create a colour shortcut

    >>> make_colour(fg='red')('hey')
    '\\x1b[31mhey\\x1b[0m'
    >>> make_colour(fg='red')('hey', fg='green')
    '\\x1b[32mhey\\x1b[0m'
    >>> make_colour(fg='red')('hey', bg='green')
    '\\x1b[31m\\x1b[42mhey\\x1b[0m'
    >>> make_colour(bold=True)('hey')
    '\\x1b[1mhey\\x1b[0m'
    >>> make_colour(underline=True)('hey')
    '\\x1b[4mhey\\x1b[0m'
    """
    @functools.wraps(click.style)
    def custom_style(*args, **kwargs):
        return click.style(*args, **{**colour_kwargs, **kwargs})

    return custom_style


e_error = make_colour(fg='red')
e_warn = make_colour(fg='yellow')
e_success = make_colour(fg='green')
e_unable = make_colour(fg='white')

e_star = make_colour(fg='bright_yellow')
e_value = make_colour(fg='blue', bold=True)
e_suggest = make_colour(fg='cyan', underline=True)



from .ascii_formatting import *  # noqa: F401, F403
from .backports import *  # noqa: F401, F403
from .collections_utils import *  # noqa: F401, F403
from .enums import *  # noqa: F401, F403
from .functions import *  # noqa: F401, F403
from .module_utils import *  # noqa: F401, F403
from .paths import *  # noqa: F401, F403
from .strings import *  # noqa: F401, F403
from .timing import *  # noqa: F401, F403
from .typing_utils import *  # noqa: F401, F403



from aox import utils
from aox.model import CombinedDayInfo, CombinedYearInfo, CombinedPartInfo, \
    CombinedInfo
from aox.summary.base_summary import BaseSummary, summary_registry


__all__ = ['SubmissionsSummary']


@summary_registry.register
class SubmissionsSummary(BaseSummary):
    """
    Show the stars for every challenge, along with links to the code and the AOC
    challenge webpage, in a year-horizontal, day-vertical table.
    """
    marker_prefix = "submissions"

    PART_STATUS_EMOJI_MAP = {
        CombinedPartInfo.STATUS_COMPLETE: ':star:',
        CombinedPartInfo.STATUS_FAILED: ':x:',
        CombinedPartInfo.STATUS_DID_NOT_ATTEMPT: '',
        CombinedPartInfo.STATUS_COULD_NOT_ATTEMPT:
        ':grey_exclamation:',
    }

    def generate(self, combined_info: CombinedInfo):
        table = self.get_table(combined_info)
        link_definitions = self.get_link_definitions(combined_info)

        return f"\n\n{table}\n\n{link_definitions}\n\n"

    def get_table(self, combined_info):
        years = sorted((
            year
            for year, year_info in combined_info.year_infos.items()
            if year_info.has_code or year_info.stars
        ), reverse=True)
        headers = ('',) + tuple(map(str, years))
        dividers = (' ---:',) + (':---:',) * len(years)
        year_stars = ('',) + tuple(
            self.get_submission_year_stars_text(combined_info.year_infos[year])
            for year in years
        )
        # noinspection PyTypeChecker
        year_links_tuples = [('',)] + [
            (
                (
                    f"[Code][{self.get_code_year_link_name(year)}]"
                    if year_info.has_code else
                    'Code'
                ),
                '&',
                '',
                f"[Challenges][ch-{str(year)[-2:]}]",
            )
            for year in years
            for year_info in (combined_info.year_infos[year],)
        ]
        # noinspection PyTypeChecker
        day_links_tuples_list = [
            [(f"{day: >2}",)] + [
                self.get_submission_year_day_stars_tuple(
                    combined_info.year_infos[year].day_infos[day])
                for year in years
            ]
            for day in range(1, 26)
        ]

        year_links, *day_links_list = \
            utils.simplify_divided_rows(
                [year_links_tuples] + day_links_tuples_list)

        table_rows = [
            headers,
            dividers,
            year_links,
            year_stars,
        ] + day_links_list

        return utils.join_rows(table_rows)

    def get_link_definitions(self, combined_info: CombinedInfo):
        years = sorted((
            year
            for year, year_info in combined_info.year_infos.items()
            if year_info.has_code or year_info.stars
        ), reverse=True)

        return "\n\n".join(
            "\n".join([
                f"[{self.get_challenge_year_link_name(year)}]: "
                f"{year_info.get_year_url()}",
                f"[{self.get_code_year_link_name(year)}]: "
                f"{year_info.relative_path}",
            ] + sum((
                [
                    f"[{self.get_challenge_day_link_name(year, day)}]: "
                    f"{day_info.get_day_url()}",
                    f"[{self.get_code_day_link_name(year, day)}]: "
                    f"{day_info.relative_path}",
                ]
                for day, day_info in (
                    (day, year_info.get_day(day))
                    for day in range(1, 26)
                )
            ), []))
            for year, year_info in (
                (year, combined_info.get_year(year))
                for year in years
            )
        )

    def get_submission_year_stars_text(
            self, year_info: CombinedYearInfo):
        if year_info.stars == 50:
            return f"50 :star: :star:"

        return "{} :star: / {} :x: / {} :grey_exclamation:".format(
            year_info.stars,
            year_info.counts_by_part_status[CombinedPartInfo.STATUS_FAILED],
            year_info.counts_by_part_status[
                CombinedPartInfo.STATUS_COULD_NOT_ATTEMPT],
        )

    def get_submission_year_day_stars_tuple(self, day_info: CombinedDayInfo):
        return (
            (
                f"[Code]"
                f"[{self.get_code_day_link_name(day_info.year, day_info.day)}]"
                if day_info.has_code else
                'Code'
            ),
            self.PART_STATUS_EMOJI_MAP[day_info.part_infos["a"].status],
            self.PART_STATUS_EMOJI_MAP[day_info.part_infos["b"].status],
            f"[Challenge]["
            f"{self.get_challenge_day_link_name(day_info.year, day_info.day)}]",
        )

    def get_code_year_link_name(self, year):
        """
        >>> SubmissionsSummary().get_code_year_link_name(2019)
        'co-19'
        """
        return f"co-{str(year)[-2:]}"

    def get_code_day_link_name(self, year, day):
        """
        >>> SubmissionsSummary().get_code_day_link_name(2019, 5)
        'co-19-05'
        >>> SubmissionsSummary().get_code_day_link_name(2019, 15)
        'co-19-15'
        """
        return f"co-{str(year)[-2:]}-{day:0>2}"

    def get_challenge_year_link_name(self, year):
        """
        >>> SubmissionsSummary().get_challenge_year_link_name(2019)
        'ch-19'
        """
        return f"ch-{str(year)[-2:]}"

    def get_challenge_day_link_name(self, year, day):
        """
        >>> SubmissionsSummary().get_challenge_day_link_name(2019, 5)
        'ch-19-05'
        >>> SubmissionsSummary().get_challenge_day_link_name(2019, 15)
        'ch-19-15'
        """
        return f"ch-{str(year)[-2:]}-{day:0>2}"



from dataclasses import fields
from pydoc import locate

import click

from aox.styling.shortcuts import e_error, e_suggest, e_value


ALL_SETTINGS_WARNINGS = {}


def get_warnings_for_new_instance():
    return dict(ALL_SETTINGS_WARNINGS)


def register_warnings(settings_class):
    for _field in fields(settings_class):
        warn = _field.metadata.get('warn')
        if not warn:
            continue
        module_attribute = _field.metadata["module_attribute"]
        ALL_SETTINGS_WARNINGS[_field.name] = warn, module_attribute

    return settings_class


class SettingsValidationError(Exception):
    pass


def warn_falsy_attribute(settings, name, module_attribute, value):
    if value:
        return value
    warn_attribute(settings, name, module_attribute, value)
    return None


def warn_missing_path(settings, name, module_attribute, value):
    if value and value.exists():
        return value
    warn_attribute(settings, name, module_attribute, value)
    return None


def warn_missing_instance(settings, name, module_attribute, value):
    return_value = value
    if isinstance(return_value, str):
        return_value = locate(return_value)
    if isinstance(return_value, type):
        return_value = return_value()
    if return_value:
        return return_value
    warn_attribute(settings, name, module_attribute, value)


# noinspection PyUnusedLocal
def warn_attribute(settings, name, module_attribute, value):
    if value:
        click.echo(
            f"The value '{e_value(str(value))}' for "
            f"{e_error(module_attribute)} is invalid.")
    elif settings.is_missing:
        click.echo(
            f"You haven't set {e_error(module_attribute)} - use "
            f"{e_suggest('aox init-settings')} to create your "
            f"settings file")
    else:
        click.echo(
            f"You haven't set {e_error(module_attribute)} in "
            f"{e_value('user_settings.py')}")


def error_on(warn):
    def error_missing_instance(settings, name, module_attribute, value):
        final_value = warn(settings, name, module_attribute, value)
        if final_value is None:
            raise SettingsValidationError(
                f"Cannot proceed with an invalid {module_attribute} value")
        return final_value

    return error_missing_instance
